###############################################################################
# Name: testEdFile.py                                                         #
# Purpose: Unit tests for ed_txt.py                                           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittests for EdFile"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: testEdFile.py 71338 2012-05-01 19:21:05Z CJP $"
__revision__ = "$Revision: 71338 $"

#-----------------------------------------------------------------------------#
# Imports
import os
import codecs
import types
import unittest

# Local imports
import common

# Module(s) to test
import ed_txt
import ebmlib

#-----------------------------------------------------------------------------#

class EdFileTest(unittest.TestCase):
    def setUp(self):
        # NEED this otherwise GetApp calls fail for some yet to be
        # determined reason even though an App has been created earlier
        # in runUnitTests...
        self.app = common.EdApp(False)

        self.path = common.GetDataFilePath(u'test_read_utf8.txt')
        self.file = ed_txt.EdFile(self.path)
        self.mtime = ebmlib.GetFileModTime(self.path)

        self.path_utf16 = common.GetDataFilePath(u'test_read_utf16.txt')
        self.mtime_utf16 = ebmlib.GetFileModTime(self.path_utf16)

        self.path_utf16_big = common.GetDataFilePath(u'test_read_utf16_big.txt')

        self.ipath = common.GetDataFilePath(u'image_test.png')
        self.img = ed_txt.EdFile(self.ipath)

        self.bpath = common.GetDataFilePath(u'test_read_utf8_bom.txt')
        self.utf8_bom_file = ed_txt.EdFile(self.bpath)

    def tearDown(self):
        self.file.Close()
        common.CleanTempDir()

    #---- Tests ----#
    def testClone(self):
        """Test making a copy of a file object"""
        clone = self.file.Clone()
        self.assertEquals(clone.GetPath(), self.file.GetPath())
        self.assertEquals(clone.Encoding, self.file.Encoding)
        self.assertEquals(clone.GetMagic(), self.file.GetMagic())
        self.assertEquals(clone.HasBom(), self.file.HasBom())
        self.assertEquals(clone.IsRawBytes(), self.file.IsRawBytes())
        self.assertEquals(clone.IsReadOnly(), self.file.IsReadOnly())

    def testRead(self):
        """Test reading from the file and getting the text"""
        txt = self.file.Read()
        self.assertFalse(self.file.HasBom())
        self.assertTrue(len(txt))

    def testReadNonPrintChars(self):
        """Test reading a plain text file that has a non printable
        character in it.
        """
        path = common.GetDataFilePath(u'non_print_char.txt')
        fileobj = ed_txt.EdFile(path)
        txt = fileobj.Read()
        self.assertTrue(type(txt) == types.UnicodeType)
        self.assertFalse(fileobj.IsRawBytes())
        self.assertFalse(fileobj.IsReadOnly())

    #---- Encoding Tests ----#

    def testReadUTF8Bom(self):
        """Test reading a utf8 bom file"""
        txt = self.utf8_bom_file.Read()
        self.assertTrue(self.utf8_bom_file.HasBom())
        self.assertTrue(len(txt))

    def testWriteUTF8Bom(self):
        """Test writing a file that has utf8 bom character"""
        txt = self.utf8_bom_file.Read()
        self.assertTrue(self.utf8_bom_file.HasBom())

        new_path = os.path.join(common.GetTempDir(), ebmlib.GetFileName(self.bpath))
        self.utf8_bom_file.SetPath(new_path) # write to test temp dir
        self.utf8_bom_file.Write(txt)

        # Open and verify that BOM was correctly written back
        f = open(new_path, 'rb')
        ntxt = f.read()
        f.close()
        self.assertTrue(ntxt.startswith(codecs.BOM_UTF8))

        # Ensure that BOM was only written once
        tmp = ntxt.lstrip(codecs.BOM_UTF8)
        self.assertFalse(tmp.startswith(codecs.BOM_UTF8))
        tmp = tmp.decode('utf-8')
        self.assertEquals(txt, tmp)

    def testWriteUTF16File(self):
        """Test that input and output bytes match"""
        fobj = ed_txt.EdFile(self.path_utf16)
        txt = fobj.Read()
        self.assertTrue(type(txt) == types.UnicodeType)
        self.assertTrue(fobj.Encoding in ('utf-16-le', 'utf_16_le', 'utf-16', 'utf_16'))
        self.assertFalse(fobj.HasBom()) # test file has no BOM

        # Get original raw bytes
        raw_bytes = common.GetFileContents(fobj.GetPath())
        
        # Write the unicode back out to disk
        out = common.GetTempFilePath('utf_16_output.txt')
        fobj.SetPath(out)
        self.assertFalse(fobj.HasBom()) # test file has no BOM
        fobj.Write(txt)

        # Get raw bytes that were just written
        new_bytes = common.GetFileContents(out)
        self.assertEquals(raw_bytes, new_bytes)

    def testReadUTF32Bom(self):
        """Test reading a file that has a UTF32 BOM"""
        fname = common.GetDataFilePath('test_read_utf32_bom.txt')
        fileutf32 = ed_txt.EdFile(fname)
        data = fileutf32.Read()
        self.assertTrue(fileutf32.HasBom(), "UTF-32 BOM not detected!")
        self.assertTrue(fileutf32.Encoding in ("utf-32", "utf_32"),
                        "Incorrect Encoding detected: %s" % fileutf32.Encoding)

    def testGetEncoding(self):
        """Test the encoding detection"""
        txt = self.file.Read()
        self.assertTrue(self.file.GetEncoding() == 'utf-8')
        fobj16 = ed_txt.EdFile(self.path_utf16)
        txt = fobj16.Read()
        enc = fobj16.GetEncoding() 
        self.assertTrue(enc in ('utf-16', 'utf_16', 'utf_16_le', 'utf-16-le'),
                        "Encoding Found: %s" % enc)

    def testGetExtension(self):
        """Test getting the file extension"""
        self.assertTrue(self.file.GetExtension() == 'txt')

    def testGetPath(self):
        """Test getting the files path"""
        self.assertTrue(self.file.GetPath() == self.path)

    def testGetMagic(self):
        """Test getting the magic comment"""
        self.file.Read()
        self.assertTrue(self.file.GetMagic())

        self.img.Read()
        self.assertFalse(self.img.GetMagic())

    def testGetModTime(self):
        """Test getting the files last modification time"""
        self.file.ModTime = self.mtime
        mtime = self.file.ModTime
        self.assertTrue(mtime == self.mtime, "Modtime was: " + str(mtime))

    def testHasBom(self):
        """Test checking if file has a bom marker or not"""
        self.assertFalse(self.file.HasBom(), "File has a BOM")

    def testIsRawBytes(self):
        """Test reading a file that can't be properly encoded and was
        read as raw bytes.

        """
        txt = self.file.Read()
        self.assertTrue(ebmlib.IsUnicode(txt))
        self.assertFalse(self.file.IsRawBytes())

        rpath = common.GetDataFilePath(u'embedded_nulls.txt')
        rfile = ed_txt.EdFile(rpath)
        txt = rfile.Read()
        self.assertTrue(ebmlib.IsUnicode(txt))
        self.assertFalse(rfile.IsRawBytes())

        bytes_value = self.img.Read()
        self.assertTrue(self.img.IsRawBytes())

    def testIsReadOnly(self):
        """Test if the file is read only or not"""
        self.assertFalse(self.file.IsReadOnly(), "File is readonly")
        self.assertEqual(self.file.IsReadOnly(), self.file.ReadOnly)

    def testSetEncoding(self):
        """Test setting the file objects encoding"""
        self.file.SetEncoding('latin1')
        self.assertTrue(self.file.GetEncoding() == 'latin1')

    #---- Module utility function tests ----#

    def testDecodeString(self):
        """Test decoding a string to unicode."""
        test = "test string"
        self.assertTrue(isinstance(test, str), "Not a string")
        uni = ed_txt.DecodeString(test, 'utf-8')
        self.assertTrue(isinstance(uni, types.UnicodeType), "Failed decode")




###############################################################################
# Name: txtutil.py                                                            #
# Purpose: Text Utilities.                                                    #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Editra Business Model Library: Text Utilities

Utility functions for managing and working with text.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: txtutil.py 67991 2011-06-20 23:48:01Z CJP $"
__revision__ = "$Revision: 67991 $"

__all__ = [ 'IsUnicode', 'DecodeString']

#-----------------------------------------------------------------------------#
# Imports
import types

#-----------------------------------------------------------------------------#

def IsUnicode(txt):
    """Is the given string a unicode string
    @param txt: object
    @return: bool

    """
    return isinstance(txt, types.UnicodeType)

def DecodeString(txt, enc):
    """Decode the given string with the given encoding,
    only attempts to decode if the given txt is not already Unicode
    @param txt: string
    @param enc: encoding 'utf-8'
    @return: unicode

    """
    if IsUnicode(txt):
        txt = txt.decode(enc)
    return txt



###############################################################################
# Name: vhdl.py                                                               #
# Purpose: Define VHDL syntax for highlighting and other features             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: vhdl.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for VHDL. Very High Scale Integrated
          Circuit Hardware Description Language
@todo: Maybe add highlighting for values S0S, S1S, ect..

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _vhdl.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# VHDL Keywords
VHDL_KW = (0, "access after alias all assert architecture array attribute for "
              "begin block body buffer bus case component configuration inout "
              "constant if else disconnect downto elsif end entity exit file "
              "function generate generic group guarded impure in inertial is "
              "label library linkage literal loop map new next null of on open "
              "others out package port postponed procedure process pure range "
              "record register reject report return select severity signal use "
              "shared subtype then to transport type unaffected units until "
              "variable wait when while with note warning error failure true "
              "false")

# VHDL Operators
VHDL_OP = (1, "and nand or nor xor xnor rol ror sla sll sra srl mod rem abs "
              "not ")

# VHDL Attributes
VHDL_AT = (2, "'high 'left 'length 'low 'range 'reverse_range 'right 'foreign "
              "'ascending 'behavior 'structure 'simple_name 'instance_name "
              "'path_name 'active 'delayed 'event 'last_active 'last_event "
              "'last_value 'quiet 'stable 'transaction 'driving 'driving_value "
              "'base 'high 'left 'leftof 'low 'pos 'pred 'rightof 'succ 'val "
              "'image 'value")

# Standard Functions
VHDL_STDF = (3, "now readline read writeline write endfile to_stdulogicvector "
                "to_bitvector to_stdulogic to_stdlogicvector resolved to_bit  "
                "to_x01 to_x01z to_UX01 rising_edge falling_edge is_x "
                "shift_right rotate_left rotate_right resize to_integer "
                "to_unsigned to_signed std_match to_01 shift_left ")

# Standard Packages
VHDL_STDP = (4, "std ieee work standard textio std_logic_1164 std_logic_arith "
                "std_logic_misc std_logic_signed std_logic_textio "
                "numeric_bit numeric_std math_complex math_real "
                "vital_timing std_logic_unsigned vital_primitives ")

# Standard Types
VHDL_STDT = (5, "bit bit_vector character boolean integer real time string "
                "severity_level positive natural signed unsigned line text "
                "std_logic std_logic_vector std_ulogic std_ulogic_vector "
                "qsim_state qsim_state_vector qsim_12state qsim_12state_vector "
                "qsim_strength mux_bit mux_vector reg_bit reg_vector wor_bit "
                "wor_vector")

# User Words
VHDL_UKW = (6, "")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_VHDL_DEFAULT, 'default_style'),
                 (stc.STC_VHDL_ATTRIBUTE, 'default_style'), # NEEDS STYLE
                 (stc.STC_VHDL_COMMENT, 'comment_style'),
                 (stc.STC_VHDL_COMMENTLINEBANG, 'comment_style'),
                 (stc.STC_VHDL_IDENTIFIER, 'default_style'),
                 (stc.STC_VHDL_KEYWORD, 'keyword_style'),
                 (stc.STC_VHDL_NUMBER, 'default_style'),
                 (stc.STC_VHDL_OPERATOR, 'operator_style'),
                 (stc.STC_VHDL_STDFUNCTION, 'funct_style'),
                 (stc.STC_VHDL_STDOPERATOR, 'operator_style'),
                 (stc.STC_VHDL_STDPACKAGE, 'pre_style'),
                 (stc.STC_VHDL_STDTYPE, 'class_style'),
                 (stc.STC_VHDL_STRING, 'string_style'),
                 (stc.STC_VHDL_STRINGEOL, 'stringeol_style'),
                 (stc.STC_VHDL_USERWORD, 'default_style') ]

#---- Extra Property Specifications ----#
FOLD = ("fold", "1")
FLD_COMMENT = ("fold.comment", "1")
FLD_COMMPACT = ("fold.compact", "1")
FLD_ATELSE  = ("fold.at.else", "1")
FLD_ATBEGIN = ("fold.at.Begin", "1")
FLD_ATPAREN = ("fold.at.Parenthese", "1")

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for VHDL""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_VHDL)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [VHDL_KW, VHDL_AT, VHDL_STDF, VHDL_STDP, VHDL_STDT]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'--']



###############################################################################
# Name: ada.py                                                                #
# Purpose: Define Ada syntax for highlighting and other features              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
 FILE: ada.py
 AUTHOR: Cody Precord
 @summary: Lexer configuration module for ada
 @todo: styles, keywords, testing

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _ada.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
ADA_KEYWORDS = (0, "abort abstract accept access aliased all array at begin "
                    "body case constant declare delay delta digits do else "
                    "elsif end entry exception exit for function generic goto "
                    "if in is limited loop new null of others out package "
                    "pragma private procedure protected raise range record "
                    "renames requeue return reverse select separate subtype "
                    "tagged task terminate then type until use when while with")

#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_ADA_CHARACTER, 'char_style'),
                 (stc.STC_ADA_CHARACTEREOL, 'stringeol_style'),
                 (stc.STC_ADA_COMMENTLINE, 'comment_style'),
                 (stc.STC_ADA_DEFAULT, 'default_style'),
                 (stc.STC_ADA_DELIMITER, 'operator_style'),
                 (stc.STC_ADA_IDENTIFIER, 'default_style'),
                 (stc.STC_ADA_ILLEGAL, 'error_style'),
                 (stc.STC_ADA_LABEL, 'keyword2_style'),   # Style This
                 (stc.STC_ADA_NUMBER, 'number_style'),
                 (stc.STC_ADA_STRING, 'string_style'),
                 (stc.STC_ADA_STRINGEOL, 'stringeol_style'),
                 (stc.STC_ADA_WORD, 'keyword_style')]

#---- Extra Properties ----#

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Ada""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_ADA)

    def GetKeywords(self):
        """Returns Specified Keywords List"""
        return [ADA_KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications"""
        return SYNTAX_ITEMS

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code"""
        return [ u'--' ]



###############################################################################
# Name: erlang.py                                                             #
# Purpose: Define Erlang syntax for highlighting and other features           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: erlang.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for the Erlang Programming Language
@todo: better styling

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _erlang.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
KEYWORDS = (0, "compile define else endif export file ifdef ifndef import "
               "include include_lib module record undef author copyright doc "
               "after begin case catch cond end fun if let of query receive "
               "when define record export import include include_lib else "
               "endif undef apply attribute call do in letrec module primop "
               "try")

#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_ERLANG_ATOM, 'default_style'),       # need new tag
                (stc.STC_ERLANG_CHARACTER, 'char_style'),
                (stc.STC_ERLANG_COMMENT, 'comment_style'),
                (stc.STC_ERLANG_DEFAULT, 'default_style'),
                (stc.STC_ERLANG_FUNCTION_NAME, 'funct_style'),
                (stc.STC_ERLANG_KEYWORD, 'keyword_style'),
                (stc.STC_ERLANG_MACRO, 'pre_style'),
                (stc.STC_ERLANG_NODE_NAME, 'string_style'),   # maybe change
                (stc.STC_ERLANG_NUMBER, 'number_style'),
                (stc.STC_ERLANG_OPERATOR, 'operator_style'),
                (stc.STC_ERLANG_RECORD, 'keyword2_style'),
                (stc.STC_ERLANG_STRING, 'string_style'),
                (stc.STC_ERLANG_UNKNOWN, 'unknown_style'),
                (stc.STC_ERLANG_VARIABLE, 'default_style')]   # need custom?

# Version specific 
if wx.VERSION >= (2, 9, 0, 0, ''):
    SYNTAX_ITEMS.append((stc.STC_ERLANG_ATOM_QUOTED, 'default_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_BIFS, 'default_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_COMMENT_DOC, 'dockey_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_COMMENT_DOC_MACRO, 'dockey_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_COMMENT_FUNCTION, 'comment_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_COMMENT_MODULE, 'comment_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_MACRO_QUOTED, 'default_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_MODULES, 'default_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_MODULES_ATT, 'default_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_NODE_NAME_QUOTED, 'default_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_PREPROC, 'pre_style')) # TODO
    SYNTAX_ITEMS.append((stc.STC_ERLANG_RECORD_QUOTED, 'default_style')) # TODO
else:
    SYNTAX_ITEMS.append((stc.STC_ERLANG_SEPARATOR, 'default_style')) # need style?

#---- Extra Properties ----#
FOLD = ('fold', '1')
FOLD_CMT = ('fold.comments', '1')
FOLD_KW = ('fold.keywords', '1')
FOLD_BRACE = ('fold.braces', '1')

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Erlang""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_ERLANG)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'%%']



###############################################################################
# Name: htmlcomp.py                                                           #
# Purpose: Simple input assistant for html and xml.                           #
# Author: Cody Precord                                                        #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Simple autocompletion support for HTML and XML documents.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__cvsid__ = "$Id: htmlcomp.py 72389 2012-08-28 16:53:09Z CJP $"
__revision__ = "$Revision: 72389 $"

#--------------------------------------------------------------------------#
# Imports
import re
import wx
import wx.stc

# Local Imports
import completer

#--------------------------------------------------------------------------#
# Standard Html Tags
TAGS = ['!--', 'a', 'abbr', 'accept', 'accesskey', 'acronym', 'action',
        'address', 'align', 'alink', 'alt', 'applet', 'archive', 'area',
        'article', 'aside', 'audio', 'axis', 'b', 'background', 'base',
        'basefont', 'bdo', 'bgcolor', 'big', 'blockquote', 'body', 'border',
        'bordercolor', 'br', 'button', 'canvas', 'caption', 'cellpadding',
        'cellspacing', 'center', 'char', 'charoff', 'charset', 'checked',
        'cite', 'cite', 'class', 'classid', 'clear', 'code', 'codebase',
        'codetype', 'col', 'colgroup', 'color', 'cols', 'colspan', 'command',
        'compact', 'content', 'coords', 'data', 'datetime', 'datalist', 'dd',
        'declare', 'defer', 'del', 'details', 'dfn', 'dialog', 'dir', 'dir',
        'disabled', 'div', 'dl', 'dt', 'dtml-call', 'dtml-comment', 'dtml-if',
        'dtml-in', 'dtml-let', 'dtml-raise', 'dtml-tree', 'dtml-try',
        'dtml-unless', 'dtml-var', 'dtml-with', 'em', 'embed', 'enctype',
        'face', 'fieldset', 'figcaption', 'figure', 'font', 'for', 'form',
        'footer', 'frame', 'gutter', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'head',
        'header', 'headers', 'height', 'hgroup', 'hr', 'href', 'hreflang',
        'hspace', 'html', 'http-equiv', 'i', 'id', 'iframe', 'img', 'input',
        'ins', 'isindex', 'ismap', 'kbd', 'keygen', 'label', 'lang', 'language',
        'legend', 'li', 'link', 'link', 'longdesc', 'lowsrc', 'map',
        'marginheight', 'marginwidth', 'mark', 'maxlength', 'menu', 'meta',
        'meter', 'method', 'multiple', 'name', 'nav', 'nohref', 'noscript',
        'nowrap', 'object', 'ol', 'optgroup', 'option', 'output', 'p', 'param',
        'pre', 'profile', 'progress', 'prompt', 'q', 'readonly', 'rel', 'rev',
        'rows', 'rowspan', 'rp', 'rt', 'ruby', 'rules', 's', 'samp', 'scheme',
        'scope', 'script', 'scrolling', 'section', 'select', 'selected',
        'shape', 'size', 'small', 'source', 'span', 'src', 'standby', 'start',
        'strike', 'strong', 'style', 'sub', 'summary', 'sup', 'tabindex',
        'table', 'target', 'tbody', 'td', 'text', 'textarea', 'tfoot', 'th',
        'thead', 'time', 'title', 'tr', 'tt', 'type', 'u', 'ul', 'url',
        'usemap', 'valign', 'value', 'valuetype', 'var', 'version', 'video',
        'vlink', 'vspace', 'width', 'wrap', 'xmp']

# Tags that usually have a new line inbetween them
NLINE_TAGS = ('body', 'head', 'html', 'ol', 'style', 'table', 'tbody', 'ul')

TAG_RE = re.compile("\<\s*([a-zA-Z][a-zA-Z0-9]*)")

PHP_AREA = [wx.stc.STC_HPHP_COMMENT, wx.stc.STC_HPHP_COMMENTLINE,
            wx.stc.STC_HPHP_COMPLEX_VARIABLE, wx.stc.STC_HPHP_DEFAULT,
            wx.stc.STC_HPHP_HSTRING, wx.stc.STC_HPHP_HSTRING_VARIABLE,
            wx.stc.STC_HPHP_NUMBER, wx.stc.STC_HPHP_OPERATOR,
            wx.stc.STC_HPHP_SIMPLESTRING,
            wx.stc.STC_HPHP_VARIABLE, wx.stc.STC_HPHP_WORD]

HTML_AREA = [wx.stc.STC_H_ASP, wx.stc.STC_H_ASPAT, wx.stc.STC_H_ATTRIBUTE,
             wx.stc.STC_H_ATTRIBUTEUNKNOWN, wx.stc.STC_H_CDATA,
             wx.stc.STC_H_COMMENT, wx.stc.STC_H_DEFAULT,
             wx.stc.STC_H_DOUBLESTRING, wx.stc.STC_H_ENTITY,
             wx.stc.STC_H_NUMBER, wx.stc.STC_H_OTHER, wx.stc.STC_H_QUESTION,
             wx.stc.STC_H_SCRIPT, wx.stc.STC_H_SGML_1ST_PARAM,
             wx.stc.STC_H_SGML_1ST_PARAM_COMMENT,
             wx.stc.STC_H_SGML_BLOCK_DEFAULT, wx.stc.STC_H_SGML_COMMAND,
             wx.stc.STC_H_SGML_COMMENT, wx.stc.STC_H_SGML_DEFAULT,
             wx.stc.STC_H_SGML_DOUBLESTRING, wx.stc.STC_H_SGML_ENTITY,
             wx.stc.STC_H_SGML_ERROR, wx.stc.STC_H_SGML_SIMPLESTRING,
             wx.stc.STC_H_SGML_SPECIAL, wx.stc.STC_H_SINGLESTRING,
             wx.stc.STC_H_TAG, wx.stc.STC_H_TAGEND,
             wx.stc.STC_H_TAGUNKNOWN, wx.stc.STC_H_VALUE,
             wx.stc.STC_H_XCCOMMENT, wx.stc.STC_H_XMLEND,
             wx.stc.STC_H_XMLSTART]

#--------------------------------------------------------------------------#

class Completer(completer.BaseCompleter):
    """HTML/XML Code completion provider"""
    def __init__(self, stc_buffer):
        super(Completer, self).__init__(stc_buffer)

        # Setup
        self.SetAutoCompKeys([ord('>'), ord('<')])
        self.SetAutoCompStops(' ')
        self.SetAutoCompFillups('')

    def GetAutoCompList(self, command):
        """Returns the list of possible completions for a
        command string.
        @param command: command lookup is done on

        """
        if command in [None, u'', u'<']:
            return list()

        buff = self.GetBuffer()
        cpos = buff.GetCurrentPos()

        # Check if we are in a php region or not
        if buff.GetStyleAt(cpos) not in HTML_AREA:
            return list()

        # Get current context
        cline = buff.GetCurrentLine()
        ccol = buff.GetColumn(cpos)
        tmp = buff.GetLine(cline).rstrip()
        if ccol < len(tmp):
            tmp = tmp[:ccol].rstrip()

        # Check if we are completing an open tag (i.e < was typed)
        if tmp.endswith('<'):
            if buff.GetLexer() == wx.stc.STC_LEX_XML:
                taglst = _FindXmlTags(buff.GetText())
            else:
                taglst = TAGS
            return completer.CreateSymbols(taglst, completer.TYPE_ELEMENT)

        # Check for a self closing tag (i.e />)
        endchk = tmp.strip().replace(u" ", u"").replace(u"\t", u"")
        if endchk.endswith(u"/>"):
            return list()

        # Try to autocomplete a closing tag (if necessary)
        tmp = tmp.rstrip('>').rstrip()
        if len(tmp) and (tmp[-1] in '"\' \t' or tmp[-1].isalpha()):
            # Walk backwards from the current line
            for line in range(cline, -1, -1):
                txt = buff.GetLine(line)
                if line == cline:
                    txt = txt[:buff.GetColumn(cpos)]

                idx = txt.rfind('<')
                if idx != -1:
                    parts = txt[idx:].lstrip('<').strip().split()
                    if len(parts):
                        tag = parts[0].rstrip('>')
                        if len(tag) and \
                           tag not in ('img', 'br', '?php', '?xml', '?') and \
                           not tag[0] in ('!', '/'):
                            rtag = u"</" + tag + u">"

                            if not parts[-1].endswith('>'):
                                rtag = u">" + rtag
                            return [ completer.Symbol(rtag, completer.TYPE_ELEMENT) ]
                    break

        return list()

    def OnCompletionInserted(self, pos, text):
        """Handle adjusting caret position after some insertions.
        @param pos: position caret was at before insertion
        @param text: text that was inserted

        """
        buff = self.GetBuffer()
        if text.strip().startswith(u"</"):
            buff.SetCurrentPos(pos) # move caret back between the tags
            # HACK: SetCurrentPos causes text to be selected
            buff.SetSelection(pos, pos)

#--------------------------------------------------------------------------#

def _FindXmlTags(text):
    """Dynamically generate a list of possible xml tags based on tags found in
    the given text.
    @param text: string
    @return: sorted list

    """
    matches = TAG_RE.findall(text)
    if len(matches):
        matches.append(u'!--')
        matches = list(set(matches))
        matches.sort()
    else:
        matches = [u'!--', ]
    return matches



###############################################################################
# Name: testExtensionRegister.py                                              #
# Purpose: Unit tests for syntax.synextreg.ExtensionRegister                  #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing the file type extension register"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: testExtensionRegister.py 62331 2009-10-08 13:42:12Z CJP $"
__revision__ = "$Revision: 62331 $"

#-----------------------------------------------------------------------------#
# Imports
import unittest

# Local modules
import common

# Module to test
import syntax.synextreg as synreg

#-----------------------------------------------------------------------------#
# Test Class

class ExtensionRegisterTest(unittest.TestCase):
    def setUp(self):
        self.reg = synreg.ExtensionRegister()

    def tearDown(self):
        self.reg.LoadDefault()

    def testAssociate(self):
        """Test that Associate successfully adds new extensions to the
        current associations of a filetype.

        """
        curr = list(self.reg[synreg.LANG_XML]) # force list copy
        self.reg.Associate(synreg.LANG_XML, 'myXml')
        self.assertTrue('myXml' in self.reg[synreg.LANG_XML])
        curr.append('myXml')
        curr.sort()
        self.assertEquals(self.reg[synreg.LANG_XML], curr)

        # Test against adding duplicates
        self.assertTrue('xml' in self.reg[synreg.LANG_XML])
        self.reg.Associate(synreg.LANG_XML, 'xml')
        self.assertEquals(self.reg[synreg.LANG_XML].count('xml'), 1)

    def testDisassociate(self):
        """Test removing an association"""
        self.assertTrue('html' in self.reg[synreg.LANG_HTML])
        self.reg.Disassociate(synreg.LANG_HTML, 'html')
        self.assertTrue('html' not in self.reg[synreg.LANG_HTML])

    def testSingleton(self):
        """Test that the syntax register is a singleton"""
        self.assertTrue(self.reg is synreg.ExtensionRegister())

    def testFileTypeFromExt(self):
        """Test getting the filetype identifier from a given extension"""
        self.assertTrue('cpp' in self.reg[synreg.LANG_CPP])
        self.assertEquals(self.reg.FileTypeFromExt('cpp'), synreg.LANG_CPP)

        # Test that non-existant returns LANG_TEXT
        self.assertEquals(self.reg.FileTypeFromExt('fake_Exent'), synreg.LANG_TXT)

    def testGetAllExtensions(self):
        self.assertTrue(isinstance(self.reg.GetAllExtensions(), list))

    def testLoadDefault(self):
        """Test loading the default settings"""
        d = list(self.reg[synreg.LANG_C]) # force list copy
        self.reg.Associate(synreg.LANG_C, 'NEW_C')
        self.assertTrue('NEW_C' in self.reg[synreg.LANG_C])
        self.assertNotEquals(d, self.reg[synreg.LANG_C])
        self.reg.LoadDefault()
        self.assertEquals(d, self.reg[synreg.LANG_C])

    # TODO:
#    def testLoadFromConfig(self):
#        pass
    
    def testSetAssociation(self):
        d = self.reg[synreg.LANG_PHP]
        self.reg.SetAssociation(synreg.LANG_PHP, "xxx yyy zzz")
        n = self.reg[synreg.LANG_PHP]
        self.assertTrue('xxx' in n)
        self.assertTrue('yyy' in n)
        self.assertTrue('zzz' in n)

        for e in d:
            self.assertFalse(e in n)

    def testGetFileExtension(self):
        """Test the module function GetFileExtensions"""
        ext = self.reg.GetAllExtensions()
        self.assertTrue(isinstance(ext, list))
        for item in ext:
            self.assertTrue(isinstance(item, basestring))



"""

  dexml:  a dead-simple Object-XML mapper for Python

Let's face it: xml is a fact of modern life.  I'd even go so far as to say
that it's *good* at what is does.  But that doesn't mean it's easy to work
with and it doesn't mean that we have to like it.  Most of the time, XML
just needs to get the hell out of the way and let you do some actual work
instead of writing code to traverse and manipulate yet another DOM.

The dexml module takes the obvious mapping between XML tags and Python objects
and lets you capture that as cleanly as possible.  Loosely inspired by Django's
ORM, you write simple class definitions to define the expected structure of
your XML document.  Like so:

  >>> import dexml
  >>> from dexml import fields
  >>> class Person(dexml.Model):
  ...   name = fields.String()
  ...   age = fields.Integer(tagname='age')

Then you can parse an XML document into an object like this:

  >>> p = Person.parse("<Person name='Foo McBar'><age>42</age></Person>")
  >>> p.name
  u'Foo McBar'
  >>> p.age
  42

And you can render an object into an XML document like this:

  >>> p = Person(name="Handsome B. Wonderful",age=36)
  >>> p.render()
  '<?xml version="1.0" ?><Person name="Handsome B. Wonderful"><age>36</age></Person>'

Malformed documents will raise a ParseError:

  >>> p = Person.parse("<Person><age>92</age></Person>")
  Traceback (most recent call last):
      ...
  ParseError: required field not found: 'name'

Of course, it gets more interesting when you nest Model definitions, like this:

  >>> class Group(dexml.Model):
  ...   name = fields.String(attrname="name")
  ...   members = fields.List(Person)
  ...
  >>> g = Group(name="Monty Python")
  >>> g.members.append(Person(name="John Cleese",age=69))
  >>> g.members.append(Person(name="Terry Jones",age=67))
  >>> g.render(fragment=True)
  '<Group name="Monty Python"><Person name="John Cleese"><age>69</age></Person><Person name="Terry Jones"><age>67</age></Person></Group>'

There's support for XML namespaces, default field values, case-insensitive
parsing, and more fun stuff.  Check out the documentation on the following
classes for more details:

  :Model:  the base class for objects that map into XML
  :Field:  the base class for individual model fields
  :Meta:   meta-information about how to parse/render a model

"""

__ver_major__ = 0
__ver_minor__ = 3
__ver_patch__ = 7
__ver_sub__ = ""
__version__ = "%d.%d.%d%s" % (__ver_major__,__ver_minor__,__ver_patch__,__ver_sub__)
                              

import copy
from xml.dom import minidom

## Local Imports
import fields
from _util import *

class Model(object):
    """Base class for dexml Model objects.

    Subclasses of Model represent a concrete type of object that can parsed 
    from or rendered to an XML document.  The mapping to/from XML is controlled
    by two things:

        * attributes declared on an inner class named 'meta'
        * fields declared using instances of fields.Field

    Here's a quick example:

        class Person(dexml.Model):
            # This overrides the default tagname of 'Person'
            class meta
                tagname = "person"
            # This maps to a 'name' attributr on the <person> tag
            name = fields.String()
            # This maps to an <age> tag within the <person> tag
            age = fields.Integer(tagname='age')

    See the 'Meta' class in this module for available meta options, and the
    'fields' submodule for available field types.
    """

    __metaclass__ = ModelMetaclass
    _fields = []

    def __init__(self,**kwds):
        """Default Model constructor.

        Keyword arguments that correspond to declared fields are processed
        and assigned to that field.
        """
        for f in self._fields:
            val = kwds.get(f.field_name)
            setattr(self,f.field_name,val)

    @classmethod
    def parse(cls,xml):
        """Produce an instance of this model from some xml.

        The given xml can be a string, a readable file-like object, or
        a DOM node; we might add support for more types in the future.
        """
        self = cls()
        node = self._make_xml_node(xml)
        self.validate_xml_node(node)
        #  Keep track of fields that have successfully parsed something
        fields_found = []
        #  Try to consume all the node's attributes
        attrs = node.attributes.values()
        for field in self._fields:
            unused_attrs = field.parse_attributes(self,attrs)
            if len(unused_attrs) < len(attrs):
                fields_found.append(field)
            attrs = unused_attrs
        for attr in attrs:
            self._handle_unparsed_node(attr)
        #  Try to consume all child nodes
        if self.meta.order_sensitive:
            self._parse_children_ordered(node,self._fields,fields_found)
        else:
            self._parse_children_unordered(node,self._fields,fields_found)
        #  Check that all required fields have been found
        for field in self._fields:
            if field.required and field not in fields_found:
                err = "required field not found: '%s'" % (field.field_name,)
                raise ParseError(err)
            field.parse_done(self)
        #  All done, return the instance so created
        return self

    def _parse_children_ordered(self,node,fields,fields_found):
        """Parse the children of the given node using strict field ordering."""
        cur_field_idx = 0 
        for child in node.childNodes:
            idx = cur_field_idx
            #  If we successfully break out of this loop, one of our
            #  fields has consumed the node.
            while idx < len(fields):
                field = fields[idx]
                res = field.parse_child_node(self,child)
                if res is PARSE_DONE:
                    if field not in fields_found:
                        fields_found.append(field)
                    cur_field_idx = idx + 1
                    break
                if res is PARSE_MORE:
                    if field not in fields_found:
                        fields_found.append(field)
                    cur_field_idx = idx
                    break
                if res is PARSE_CHILDREN:
                    self._parse_children_ordered(child,[field],fields_found)
                    cur_field_idx = idx
                    break
                idx += 1
            else:
                self._handle_unparsed_node(child)

    def _parse_children_unordered(self,node,fields,fields_found):
        """Parse the children of the given node using loose field ordering."""
        done_fields = {}
        for child in node.childNodes:
            idx = 0
            #  If we successfully break out of this loop, one of our
            #  fields has consumed the node.
            while idx < len(fields):
                if idx in done_fields:
                    idx += 1
                    continue
                field = fields[idx]
                res = field.parse_child_node(self,child)
                if res is PARSE_DONE:
                    done_fields[idx] = True
                    if field not in fields_found:
                        fields_found.append(field)
                    break
                if res is PARSE_MORE:
                    if field not in fields_found:
                        fields_found.append(field)
                    break
                if res is PARSE_CHILDREN:
                    self._parse_children_unordered(child,[field],fields_found)
                    break
                idx += 1
            else:
                self._handle_unparsed_node(child)

    def _handle_unparsed_node(self,node):
        if not self.meta.ignore_unknown_elements:
            if node.nodeType == node.ELEMENT_NODE:
                err = "unknown element: %s" % (node.nodeName,)
                raise ParseError(err)
            elif node.nodeType in (node.TEXT_NODE,node.CDATA_SECTION_NODE):
                if node.nodeValue.strip():
                    err = "unparsed text node: %s" % (node.nodeValue,)
                    raise ParseError(err)
            elif node.nodeType == node.ATTRIBUTE_NODE:
                if not node.nodeName.startswith("xml"):
                    err = "unknown attribute: %s" % (node.name,)
                    raise ParseError(err)

    def render(self,encoding=None,fragment=False,nsmap=None):
        """Produce XML from this model's instance data.

        A unicode string will be returned if any of the objects contain
        unicode values; specifying the 'encoding' argument forces generation
        of an ASCII string.

        By default a complete XML document is produced, including the
        leading "<?xml>" declaration.  To generate an XML fragment set
        the 'fragment' argument to True.

        The 'nsmap' argument maintains the current stack of namespace
        prefixes used during rendering; it maps each prefix to a list of
        namespaces, with the first item in the list being the current
        namespace for that prefix.  This argument should never be given
        directly; it is for internal use by the rendering routines.
         
        """
        if nsmap is None:
            nsmap = {}
        data = []
        if not fragment:
            if encoding:
                s = '<?xml version="1.0" encoding="%s" ?>' % (encoding,)
                data.append(s)
            else:
                data.append('<?xml version="1.0" ?>')
        data.extend(self._render(nsmap))
        xml = "".join(data)
        if encoding:
            xml = xml.encode(encoding)
        return xml

    def _render(self,nsmap):
        """Render this model as an XML fragment."""
        #  Determine opening and closing tags
        pushed_ns = False
        if self.meta.namespace:
            namespace = self.meta.namespace
            prefix = self.meta.namespace_prefix
            try:
                cur_ns = nsmap[prefix]
            except KeyError:
                cur_ns = []
                nsmap[prefix] = cur_ns
            if prefix:
                tagname = "%s:%s" % (prefix,self.meta.tagname)
                open_tag_contents = [tagname]
                if not cur_ns or cur_ns[0] != namespace:
                    cur_ns.insert(0,namespace)
                    pushed_ns = True
                    open_tag_contents.append('xmlns:%s="%s"'%(prefix,namespace))
                close_tag_contents = tagname
            else:
                open_tag_contents = [self.meta.tagname]
                if not cur_ns or cur_ns[0] != namespace:
                    cur_ns.insert(0,namespace)
                    pushed_ns = True
                    open_tag_contents.append('xmlns="%s"'%(namespace,))
                close_tag_contents = self.meta.tagname
        else:
            open_tag_contents = [self.meta.tagname] 
            close_tag_contents = self.meta.tagname
        # Find the attributes and child nodes
        attrs = []
        children = []
        num = 0
        for f in self._fields:
            val = getattr(self,f.field_name)
            attrs.extend(f.render_attributes(self,val,nsmap))
            children.extend(f.render_children(self,val,nsmap))
            if len(attrs) + len(children) == num and f.required:
                raise RenderError("Field '%s' is missing" % (f.field_name,))
        #  Actually construct the XML
        if pushed_ns:
            nsmap[prefix].pop(0)
        open_tag_contents.extend(attrs)
        if children:
            yield "<%s>" % (" ".join(open_tag_contents),)
            for chld in children:
                yield chld
            yield "</%s>" % (close_tag_contents,)
        else:
            yield "<%s />" % (" ".join(open_tag_contents),)

    @staticmethod
    def _make_xml_node(xml):
        """Transform a variety of input formats to an XML DOM node."""
        try:
            ntype = xml.nodeType
        except AttributeError:
            if isinstance(xml,basestring):
                try:
                    xml = minidom.parseString(xml)
                except Exception, e:
                    raise XmlError(e)
            elif hasattr(xml,"read"):
                try:
                    xml = minidom.parse(xml)
                except Exception, e:
                    raise XmlError(e)
            else:
                raise ValueError("Can't convert that to an XML DOM node")
            node = xml.documentElement
        else:
            if ntype == xml.DOCUMENT_NODE:
                node = xml.documentElement
            else:
                node = xml
        return node

    @classmethod
    def validate_xml_node(cls,node):
        """Check that the given xml node is valid for this object.

        Here 'valid' means that it is the right tag, in the right
        namespace.  We might add more eventually...
        """
        if node.nodeType != node.ELEMENT_NODE:
            err = "Class '%s' got a non-element node"
            err = err % (cls.__name__,)
            raise ParseError(err)
        equals = (lambda a, b: a == b) if cls.meta.case_sensitive else (lambda a, b: a.lower() == b.lower())
        if not equals(node.localName, cls.meta.tagname):
            err = "Class '%s' got tag '%s' (expected '%s')"
            err = err % (cls.__name__,node.localName,
                         cls.meta.tagname)
            raise ParseError(err)
        if cls.meta.namespace:
            if node.namespaceURI != cls.meta.namespace:
                err = "Class '%s' got namespace '%s' (expected '%s')"
                err = err % (cls.__name__,node.namespaceURI,
                             cls.meta.namespace)
                raise ParseError(err)
        else:
            if node.namespaceURI:
                err = "Class '%s' got namespace '%s' (expected no namespace)"
                err = err % (cls.__name__,node.namespaceURI,)
                raise ParseError(err)


